#include <msp430.h>
#include <stdint.h>
#include <limits.h>

/*
 Entwicklungsumgebung ist ein MSP-EXP430G2 mit MSP430F2011
 Ansteuerung eines DHT11-Sensors an P1.2 per PWM und Capture-Interface
 */

#define SMCLOCK 8000000
#define DELAY18MS ((20e-3/8*SMCLOCK)+1)
#define BITTIME99us (unsigned int)(99e-6*SMCLOCK)

// 4 Byte Nutzdaten des DHT11
struct DHT11DATA {
	uint8_t humidity,humhum,temphi,templo;
} letzteMessung;

struct {
	unsigned int bitzeit:6;
	unsigned int gewandelt:1;
	unsigned int crcfehler:1;
	unsigned int overflow:1;
	unsigned int o_zeit:6;
} flags;

#define DHT11_STARTBIT 42
#define DHT11_UEBERNAHME 41
#define DHT11_ACKNOWLEDGE 40

unsigned int debugnr;

// lese_DHT11: Anstossen des Auslesens der Temperatur- und Feuchtigkeitswerte
void lese_DHT11(void)
{
	// 18ms lang Lowpegel auf der Datenleitung anlegen.
    TACTL = 0;					// Timer A2 anhalten
    P1DIR &= ~BIT2;				// whrend der Timereinstellung P1.2 hochohmig schalten
    P1SEL &= ~BIT2;
	TACCTL1 = OUTMOD_1 | CCIE;	// wollte per DMA treiben, nutze aber nur den Zeitgeber bis ISR()
	TACCR1 = DELAY18MS;
	TACTL = TASSEL_2 | ID_3 | TACLR;	// Teiler durch 8, damit 18ms erzeugbar sind.
	flags.bitzeit = DHT11_STARTBIT;
	P1OUT &= ~BIT2;
	P1DIR |= BIT2;
	TACTL |= MC_2;				// Zhler loslaufen lassen -> TIMERA1_VECTOR
}

int main(void) {
    WDTCTL = WDTPW | WDTHOLD;	// Stop watchdog timer
    DCOCTL = CALDCO_8MHZ;
    BCSCTL1 = CALBC1_8MHZ;
    flags.gewandelt = 0;
    flags.crcfehler = 0;
    flags.overflow = 0;
    debugnr = 55;
    _enable_interrupts();
    do {
    	lese_DHT11();
    	_low_power_mode_1();
    	_delay_cycles(40000);
    } while(!flags.gewandelt);
    _nop();
	return 0;
}

#pragma vector=TIMERA1_VECTOR
__interrupt void TIMERA1_ISR(void)
{
	static uint8_t empfang[5];
	static unsigned int alter_zaehler;
	unsigned int nr,diff;
	if(TAIV==TAIV_TACCR1) {
		if(TACCTL1&COV) {
			flags.overflow = 1;
			flags.o_zeit = flags.bitzeit;
//			_bic_SR_register_on_exit(SCG0|CPUOFF);
//			return;
		}
		if(flags.bitzeit<DHT11_ACKNOWLEDGE) {
			if(flags.bitzeit==debugnr) {
				_nop();
			}
			nr = flags.bitzeit>>3;
			diff = TACCR1 - alter_zaehler;
			if(diff>INT_MAX) diff+=UINT_MAX;
			alter_zaehler = TACCR1;
			empfang[nr]<<=1;
			if(diff>BITTIME99us) empfang[nr]++;
			if(flags.bitzeit) flags.bitzeit--;
			else {
				diff = (empfang[4] + empfang[3] + empfang[2] + empfang[1])&0xff;
				if(diff==empfang[0]) {
					letzteMessung.humidity = empfang[4];
					letzteMessung.humhum = empfang[3];
					letzteMessung.temphi = empfang[2];
					letzteMessung.templo = empfang[1];
					flags.gewandelt = 1;
				} else flags.crcfehler = 1;
				TACTL = 0;
				TACCR1 = 0;
				_bic_SR_register_on_exit(SCG0|CPUOFF);
			}
		} else if(flags.bitzeit==DHT11_STARTBIT) {
			// Zhler A2 umstellen auf "capture"
			TACTL = 0;
			P1DIR &= ~BIT2;
			P1SEL |= BIT2;
			TACCTL1 = CM_2 | CCIS_0 | SCS | CAP | CCIE;
			TACTL = TASSEL_2 | ID_0 | MC_2 | TACLR;
			flags.bitzeit--;
		} else if(flags.bitzeit==DHT11_UEBERNAHME) {
			flags.bitzeit--;
		} else if(flags.bitzeit==DHT11_ACKNOWLEDGE) {
			alter_zaehler = TAR;
			flags.bitzeit--;
		}
	}
}

#pragma vector=COMPARATORA_VECTOR
__interrupt void Comp_ISR(void)
{
	_nop();
}

#pragma vector=PORT1_VECTOR
__interrupt void PORT1_ISR(void)
{
	_nop();
}

#pragma vector=PORT2_VECTOR
__interrupt void PORT2_ISR(void)
{
	_nop();
}

#pragma vector=TIMERA0_VECTOR
__interrupt void TIMERA0_ISR(void)
{
	_nop();
}

#pragma vector=WDT_VECTOR
__interrupt void WDT_ISR(void)
{
	_nop();
}

#pragma vector=NMI_VECTOR
__interrupt void NMI_ISR(void)
{
	_nop();
}
